from typing import Callable, Optional, Tuple, Union

import numpy as np

from .inv import inv2


def eigh(
    A: np.ndarray, B: Optional[np.ndarray] = None, type: Optional[int] = 1
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    r"""Compute the (generalized) eigenvalues and eigenvectors of a complex Hermitian \
    (conjugate symmetric) or a real symmetric matrix.

    If ``B`` is ``None``, solve :math:`\boldsymbol{A}\boldsymbol{z} = \lambda\boldsymbol{z}`.

    If ``B`` is given,
    solve :math:`\boldsymbol{A}\boldsymbol{z} = \lambda\boldsymbol{B}\boldsymbol{z}`.

    Args:
        A (numpy.ndarray):
            A complex Hermitian matrix with shape of (\*, n_channels, n_channels).
        B (numpy.ndarray, optional):
            A complex Hermitian matrix with shape of (\*, n_channels, n_channels).
        type (int):
            For the generalized eigenproblem, this value specifies the type of problem.
            Only ``1``, ``2``, and ``3`` are supported.

            - When ``type=1``, solve :math:`\boldsymbol{Az}=\lambda\boldsymbol{Bz}`.
            - When ``type=2``, solve :math:`\boldsymbol{ABz}=\lambda\boldsymbol{z}`.
            - When ``type=3``, solve :math:`\boldsymbol{BAz}=\lambda\boldsymbol{z}`.

    Returns:
        A tuple of (eigenvalues, eigenvectors)
            - Eigenvalues have shape of (\*, n_channels).
            - Eigenvectors have shape of (\*, n_channels, n_channels).

    .. note::
        If ``B`` is given, we use cholesky decomposition to
        satisfy :math:`\boldsymbol{L}\boldsymbol{L}^{\mathsf{H}}=\boldsymbol{B}`.

        Then, solve :math:`\boldsymbol{C}\boldsymbol{y} = \lambda\boldsymbol{y}`,
        where :math:`\boldsymbol{C}=\boldsymbol{L}^{-1}\boldsymbol{A}\boldsymbol{L}^{-\mathsf{H}}`.

        The generalized eigenvalues of :math:`\boldsymbol{A}` and :math:`\boldsymbol{B}`
        are computed by :math:`\boldsymbol{L}^{-\mathsf{H}}\boldsymbol{y}`.

    Examples:

        .. code-block:: python

            >>> import numpy as np
            >>> from ssspy.linalg import eigh
            >>> A = np.array([[1, -2j], [2j, 3]])
            >>> lamb, z = eigh(A)
            >>> lamb; z
            array([-0.23606798,  4.23606798])
            array([[-0.85065081+0.j        , -0.52573111+0.j        ],
                [ 0.        +0.52573111j,  0.        -0.85065081j]])
            >>> np.allclose(A @ z, lamb * z)
            True

        .. code-block:: python

            >>> import numpy as np
            >>> from ssspy.linalg import eigh
            >>> A = np.array([[1, -2j], [2j, 3]])
            >>> B = np.array([[2, -3j], [3j, 5]])
            >>> lamb, z = eigh(A, B)
            >>> lamb; z
            array([-1.61803399,  0.61803399])
            array([[ 2.22703273+0.j        , -0.20081142+0.j        ],
                [ 0.        -1.37638192j,  0.        -0.3249197j ]])
            >>> np.allclose(A @ z, lamb * (B @ z))
            True
    """
    if B is None:
        return np.linalg.eigh(A)

    lamb, z = _eigh(A, B, type=type, inv=np.linalg.inv)

    return lamb, z


def eigh2(
    A: np.ndarray, B: Optional[np.ndarray] = None, type: Optional[int] = 1
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    r"""Compute the (generalized) eigenvalues and eigenvectors of a 2x2 complex Hermitian \
    (conjugate symmetric) or a real symmetric matrix.

    If ``B`` is ``None``, solve :math:`\boldsymbol{A}\boldsymbol{z} = \lambda\boldsymbol{z}`.

    If ``B`` is given,
    solve :math:`\boldsymbol{A}\boldsymbol{z} = \lambda\boldsymbol{B}\boldsymbol{z}`.

    Args:
        A (numpy.ndarray):
            A complex Hermitian matrix with shape of (\*, 2, 2).
        B (numpy.ndarray, optional):
            A complex Hermitian matrix with shape of (\*, 2, 2).
        type (int):
            For the generalized eigenproblem, this value specifies the type of problem.
            Only ``1``, ``2``, and ``3`` are supported.

            - When ``type=1``, solve :math:`\boldsymbol{Az}=\lambda\boldsymbol{Bz}`.
            - When ``type=2``, solve :math:`\boldsymbol{ABz}=\lambda\boldsymbol{z}`.
            - When ``type=3``, solve :math:`\boldsymbol{BAz}=\lambda\boldsymbol{z}`.

    Returns:
        A tuple of (eigenvalues, eigenvectors)
            - Eigenvalues have shape of (\*, 2).
            - Eigenvectors have shape of (\*, 2, 2).

    .. note::
        If ``B`` is given, we use cholesky decomposition to
        satisfy :math:`\boldsymbol{L}\boldsymbol{L}^{\mathsf{H}}=\boldsymbol{B}`.

        Then, solve :math:`\boldsymbol{C}\boldsymbol{y} = \lambda\boldsymbol{y}`,
        where :math:`\boldsymbol{C}=\boldsymbol{L}^{-1}\boldsymbol{A}\boldsymbol{L}^{-\mathsf{H}}`.

        The generalized eigenvalues of :math:`\boldsymbol{A}` and :math:`\boldsymbol{B}`
        are computed by :math:`\boldsymbol{L}^{-\mathsf{H}}\boldsymbol{y}`.

        See also https://github.com/tky823/ssspy/issues/115 for this implementation.

    Examples:

        .. code-block:: python

            >>> import numpy as np
            >>> from ssspy.linalg import eigh2
            >>> A = np.array([[1, -2j], [2j, 3]])
            >>> lamb, z = eigh2(A)
            >>> lamb; z
            array([-0.23606798,  4.23606798])
            array([[-0.85065081+0.j        , -0.52573111+0.j        ],
                [ 0.        +0.52573111j,  0.        -0.85065081j]])
            >>> np.allclose(A @ z, lamb * z)
            True

        .. code-block:: python

            >>> import numpy as np
            >>> from ssspy.linalg import eigh2
            >>> A = np.array([[1, -2j], [2j, 3]])
            >>> B = np.array([[2, -3j], [3j, 5]])
            >>> lamb, z = eigh2(A, B)
            >>> lamb; z
            array([-1.61803399,  0.61803399])
            array([[ 2.22703273+0.j        , -0.20081142+0.j        ],
                [ 0.        -1.37638192j,  0.        -0.3249197j ]])
            >>> np.allclose(A @ z, lamb * (B @ z))
            True
    """
    assert A.shape[-2:] == (2, 2), "2x2 matrix is expected, but given shape of {}.".format(A.shape)

    if B is None:
        return np.linalg.eigh(A)

    lamb, z = _eigh(A, B, type=type, inv=inv2)

    return lamb, z


def _eigh(
    A: np.ndarray,
    B: np.ndarray,
    type: int = 1,
    inv: Callable[[np.ndarray], np.ndarray] = None,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    if inv is None:
        inv = np.linalg.inv

    L = np.linalg.cholesky(B)

    if type == 1:
        L_inv = inv(L)
        L_inv_Hermite = np.swapaxes(L_inv, -2, -1)

        if np.iscomplexobj(L_inv_Hermite):
            L_inv_Hermite = L_inv_Hermite.conj()

        C = L_inv @ A @ L_inv_Hermite
    elif type in [2, 3]:
        L_Hermite = np.swapaxes(L, -2, -1)

        if np.iscomplexobj(L_Hermite):
            L_Hermite = L_Hermite.conj()

        C = L_Hermite @ A @ L

        if type == 2:
            L_inv_Hermite = inv(L_Hermite)
        else:
            L_inv_Hermite = None
    else:
        raise ValueError("Invalid type={} is given.".format(type))

    lamb, y = np.linalg.eigh(C)

    if type in [1, 2]:
        z = L_inv_Hermite @ y
    elif type == 3:
        z = L @ y
    else:
        raise ValueError("Invalid type={} is given.".format(type))

    return lamb, z



import os
import sys
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pytest
import scipy.signal as ss

from ssspy.bss.ilrma import GGDILRMA, TILRMA, GaussILRMA, ILRMABase

ssspy_tests_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
sys.path.append(ssspy_tests_dir)

from dummy.callback import DummyCallback, dummy_function  # noqa: E402
from dummy.utils.dataset import download_sample_speech_data  # noqa: E402

max_duration = 0.5
n_fft = 512
hop_length = 256
n_bins = n_fft // 2 + 1
n_iter = 3
rng = np.random.default_rng(42)

parameters_dof = [100]
parameters_beta = [0.5, 1.5]
parameters_spatial_algorithm = ["IP", "IP1", "IP2", "ISS", "ISS1", "ISS2", "IPA"]
parameters_source_algorithm = ["MM", "ME"]
parameters_callbacks = [None, dummy_function, [DummyCallback(), dummy_function]]
parameters_scale_restoration = [True, False, "projection_back", "minimal_distortion_principle"]
parameters_ilrma_base = [2]
parameters_ilrma_latent = [
    (
        2,
        4,
        2,
        {
            "demix_filter": np.tile(np.eye(2, dtype=np.complex128), (n_bins, 1, 1)),
            "latent": rng.random((2, 4)),
            "basis": rng.random((n_bins, 4)),
        },
    ),
    (3, 3, 1, {}),
]
parameters_ilrma_wo_latent = [
    (
        2,
        2,
        2,
        {
            "demix_filter": np.tile(np.eye(2, dtype=np.complex128), (n_bins, 1, 1)),
            "basis": rng.random((2, n_bins, 2)),
        },
    ),
    (
        3,
        1,
        1,
        {},
    ),
]
parameters_normalization_latent = [True, False, "power"]
parameters_normalization_wo_latent = [True, False, "power", "projection_back"]


@pytest.mark.parametrize(
    "n_basis",
    parameters_ilrma_base,
)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_ilrma_base(
    n_basis: int,
    callbacks: Optional[Union[Callable[[GaussILRMA], None], List[Callable[[GaussILRMA], None]]]],
    scale_restoration: Union[str, bool],
):
    ilrma = ILRMABase(
        n_basis,
        partitioning=True,
        callbacks=callbacks,
        scale_restoration=scale_restoration,
        rng=np.random.default_rng(42),
    )

    print(ilrma)


@pytest.mark.parametrize(
    "n_sources, n_basis, domain, reset_kwargs",
    parameters_ilrma_latent,
)
@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
@pytest.mark.parametrize("source_algorithm", parameters_source_algorithm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("normalization", parameters_normalization_latent)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_gauss_ilrma_latent(
    n_sources: int,
    n_basis: int,
    spatial_algorithm: str,
    source_algorithm: str,
    domain: float,
    callbacks: Optional[Union[Callable[[GaussILRMA], None], List[Callable[[GaussILRMA], None]]]],
    normalization: Optional[Union[str, bool]],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[str, Any],
):
    if spatial_algorithm in ["IP", "ISS"] and not pytest.run_redundant:
        pytest.skip(reason="Need --run-redundant option to run.")

    if n_sources < 4:
        sisec2010_tag = "dev1_female3"
    elif n_sources == 4:
        sisec2010_tag = "dev1_female4"
    else:
        raise ValueError("n_sources should be less than 5.")

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    kwargs = {
        "spatial_algorithm": spatial_algorithm,
        "source_algorithm": source_algorithm,
        "domain": domain,
        "partitioning": True,
        "callbacks": callbacks,
        "normalization": normalization,
        "scale_restoration": scale_restoration,
        "rng": np.random.default_rng(42),
    }

    if source_algorithm == "ME" and domain != 2:
        with pytest.raises(AssertionError) as e:
            ilrma = GaussILRMA(n_basis, **kwargs)

        assert str(e.value) == "domain parameter should be 2 when you specify ME algorithm."
    else:
        ilrma = GaussILRMA(n_basis, **kwargs)
        spectrogram_est = ilrma(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

        assert spectrogram_mix.shape == spectrogram_est.shape
        assert type(ilrma.loss[-1]) is float

        if spatial_algorithm in ["ISS", "ISS1", "ISS2"]:
            assert ilrma.demix_filter is None

        print(ilrma)


@pytest.mark.parametrize(
    "n_sources, n_basis, domain, reset_kwargs",
    parameters_ilrma_wo_latent,
)
@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
@pytest.mark.parametrize("source_algorithm", parameters_source_algorithm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("normalization", parameters_normalization_wo_latent)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_gauss_ilrma_wo_latent(
    n_sources: int,
    n_basis: int,
    spatial_algorithm: str,
    source_algorithm: str,
    domain: float,
    callbacks: Optional[Union[Callable[[GaussILRMA], None], List[Callable[[GaussILRMA], None]]]],
    normalization: Optional[Union[str, bool]],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[str, Any],
):
    if spatial_algorithm in ["IP", "ISS"] and not pytest.run_redundant:
        pytest.skip(reason="Need --run-redundant option to run.")

    if n_sources < 4:
        sisec2010_tag = "dev1_female3"
    elif n_sources == 4:
        sisec2010_tag = "dev1_female4"
    else:
        raise ValueError("n_sources should be less than 5.")

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    kwargs = {
        "spatial_algorithm": spatial_algorithm,
        "source_algorithm": source_algorithm,
        "domain": domain,
        "partitioning": False,
        "callbacks": callbacks,
        "normalization": normalization,
        "scale_restoration": scale_restoration,
        "rng": np.random.default_rng(42),
    }

    if source_algorithm == "ME" and domain != 2:
        with pytest.raises(AssertionError) as e:
            ilrma = GaussILRMA(n_basis, **kwargs)

        assert str(e.value) == "domain parameter should be 2 when you specify ME algorithm."
    else:
        ilrma = GaussILRMA(n_basis, **kwargs)
        spectrogram_est = ilrma(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

        assert spectrogram_mix.shape == spectrogram_est.shape
        assert type(ilrma.loss[-1]) is float

        if spatial_algorithm in ["ISS", "ISS1", "ISS2"]:
            assert ilrma.demix_filter is None

        print(ilrma)


@pytest.mark.parametrize(
    "n_sources, n_basis, domain, reset_kwargs",
    parameters_ilrma_latent,
)
@pytest.mark.parametrize("dof", parameters_dof)
@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
@pytest.mark.parametrize("source_algorithm", parameters_source_algorithm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("normalization", parameters_normalization_latent)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_t_ilrma_latent(
    n_sources: int,
    n_basis: int,
    dof: float,
    spatial_algorithm: str,
    source_algorithm: str,
    domain: float,
    callbacks: Optional[Union[Callable[[GaussILRMA], None], List[Callable[[GaussILRMA], None]]]],
    normalization: Optional[Union[str, bool]],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[str, Any],
):
    if spatial_algorithm in ["IP", "ISS"] and not pytest.run_redundant:
        pytest.skip(reason="Need --run-redundant option to run.")

    if n_sources < 4:
        sisec2010_tag = "dev1_female3"
    elif n_sources == 4:
        sisec2010_tag = "dev1_female4"
    else:
        raise ValueError("n_sources should be less than 5.")

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    kwargs = {
        "dof": dof,
        "spatial_algorithm": spatial_algorithm,
        "source_algorithm": source_algorithm,
        "domain": domain,
        "partitioning": True,
        "callbacks": callbacks,
        "normalization": normalization,
        "scale_restoration": scale_restoration,
        "rng": np.random.default_rng(42),
    }

    if spatial_algorithm == "IPA":
        with pytest.raises(ValueError) as e:
            ilrma = TILRMA(n_basis, **kwargs)

        assert str(e.value) == "IPA is not supported for t-ILRMA."
    elif source_algorithm == "ME" and domain != 2:
        with pytest.raises(AssertionError) as e:
            ilrma = TILRMA(n_basis, **kwargs)

        assert str(e.value) == "domain parameter should be 2 when you specify ME algorithm."
    else:
        ilrma = TILRMA(n_basis, **kwargs)
        spectrogram_est = ilrma(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

        assert spectrogram_mix.shape == spectrogram_est.shape
        assert type(ilrma.loss[-1]) is float

        if spatial_algorithm in ["ISS", "ISS1", "ISS2"]:
            assert ilrma.demix_filter is None

        print(ilrma)


@pytest.mark.parametrize(
    "n_sources, n_basis, domain, reset_kwargs",
    parameters_ilrma_wo_latent,
)
@pytest.mark.parametrize("dof", parameters_dof)
@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
@pytest.mark.parametrize("source_algorithm", parameters_source_algorithm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("normalization", parameters_normalization_wo_latent)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_t_ilrma_wo_latent(
    n_sources: int,
    n_basis: int,
    dof: float,
    spatial_algorithm: str,
    source_algorithm: str,
    domain: float,
    callbacks: Optional[Union[Callable[[GaussILRMA], None], List[Callable[[GaussILRMA], None]]]],
    normalization: Optional[Union[str, bool]],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[str, Any],
):
    if spatial_algorithm in ["IP", "ISS"] and not pytest.run_redundant:
        pytest.skip(reason="Need --run-redundant option to run.")

    if n_sources < 4:
        sisec2010_tag = "dev1_female3"
    elif n_sources == 4:
        sisec2010_tag = "dev1_female4"
    else:
        raise ValueError("n_sources should be less than 5.")

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    kwargs = {
        "dof": dof,
        "spatial_algorithm": spatial_algorithm,
        "source_algorithm": source_algorithm,
        "domain": domain,
        "partitioning": False,
        "callbacks": callbacks,
        "normalization": normalization,
        "scale_restoration": scale_restoration,
        "rng": np.random.default_rng(42),
    }

    if spatial_algorithm == "IPA":
        with pytest.raises(ValueError) as e:
            ilrma = TILRMA(n_basis, **kwargs)

        assert str(e.value) == "IPA is not supported for t-ILRMA."
    elif source_algorithm == "ME" and domain != 2:
        with pytest.raises(AssertionError) as e:
            ilrma = TILRMA(n_basis, **kwargs)

        assert str(e.value) == "domain parameter should be 2 when you specify ME algorithm."
    else:
        ilrma = TILRMA(n_basis, **kwargs)
        spectrogram_est = ilrma(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

        assert spectrogram_mix.shape == spectrogram_est.shape
        assert type(ilrma.loss[-1]) is float

        if spatial_algorithm in ["ISS", "ISS1", "ISS2"]:
            assert ilrma.demix_filter is None

        print(ilrma)


@pytest.mark.parametrize(
    "n_sources, n_basis, domain, reset_kwargs",
    parameters_ilrma_latent,
)
@pytest.mark.parametrize("beta", parameters_beta)
@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
@pytest.mark.parametrize("source_algorithm", parameters_source_algorithm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("normalization", parameters_normalization_latent)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_ggd_ilrma_latent(
    n_sources: int,
    n_basis: int,
    beta: float,
    spatial_algorithm: str,
    source_algorithm: str,
    domain: float,
    callbacks: Optional[Union[Callable[[GaussILRMA], None], List[Callable[[GaussILRMA], None]]]],
    normalization: Optional[Union[str, bool]],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[str, Any],
):
    if spatial_algorithm in ["IP", "ISS"] and not pytest.run_redundant:
        pytest.skip(reason="Need --run-redundant option to run.")

    if n_sources < 4:
        sisec2010_tag = "dev1_female3"
    elif n_sources == 4:
        sisec2010_tag = "dev1_female4"
    else:
        raise ValueError("n_sources should be less than 5.")

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    kwargs = {
        "beta": beta,
        "spatial_algorithm": spatial_algorithm,
        "source_algorithm": source_algorithm,
        "domain": domain,
        "partitioning": True,
        "callbacks": callbacks,
        "normalization": normalization,
        "scale_restoration": scale_restoration,
        "rng": np.random.default_rng(42),
    }

    if source_algorithm == "ME":
        with pytest.raises(AssertionError) as e:
            ilrma = GGDILRMA(n_basis, **kwargs)

        assert str(e.value) == "Not support {}.".format(source_algorithm)
    elif spatial_algorithm == "IPA":
        with pytest.raises(ValueError) as e:
            ilrma = GGDILRMA(n_basis, **kwargs)

        assert str(e.value) == "IPA is not supported for GGD-ILRMA."
    else:
        ilrma = GGDILRMA(n_basis, **kwargs)
        spectrogram_est = ilrma(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

        assert spectrogram_mix.shape == spectrogram_est.shape
        assert type(ilrma.loss[-1]) is float

        if spatial_algorithm in ["ISS", "ISS1", "ISS2"]:
            assert ilrma.demix_filter is None

        print(ilrma)


@pytest.mark.parametrize(
    "n_sources, n_basis, domain, reset_kwargs",
    parameters_ilrma_wo_latent,
)
@pytest.mark.parametrize("beta", parameters_beta)
@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
@pytest.mark.parametrize("source_algorithm", parameters_source_algorithm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("normalization", parameters_normalization_wo_latent)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_ggd_ilrma_wo_latent(
    n_sources: int,
    n_basis: int,
    beta: float,
    spatial_algorithm: str,
    source_algorithm: str,
    domain: float,
    callbacks: Optional[Union[Callable[[GaussILRMA], None], List[Callable[[GaussILRMA], None]]]],
    normalization: Optional[Union[str, bool]],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[str, Any],
):
    if spatial_algorithm in ["IP", "ISS"] and not pytest.run_redundant:
        pytest.skip(reason="Need --run-redundant option to run.")

    if n_sources < 4:
        sisec2010_tag = "dev1_female3"
    elif n_sources == 4:
        sisec2010_tag = "dev1_female4"
    else:
        raise ValueError("n_sources should be less than 5.")

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    kwargs = {
        "beta": beta,
        "spatial_algorithm": spatial_algorithm,
        "source_algorithm": source_algorithm,
        "domain": domain,
        "partitioning": False,
        "callbacks": callbacks,
        "normalization": normalization,
        "scale_restoration": scale_restoration,
        "rng": np.random.default_rng(42),
    }

    if source_algorithm == "ME":
        with pytest.raises(AssertionError) as e:
            ilrma = GGDILRMA(n_basis, **kwargs)

        assert str(e.value) == "Not support {}.".format(source_algorithm)
    elif spatial_algorithm == "IPA":
        with pytest.raises(ValueError) as e:
            ilrma = GGDILRMA(n_basis, **kwargs)

        assert str(e.value) == "IPA is not supported for GGD-ILRMA."
    else:
        ilrma = GGDILRMA(n_basis, **kwargs)
        spectrogram_est = ilrma(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

        assert spectrogram_mix.shape == spectrogram_est.shape
        assert type(ilrma.loss[-1]) is float

        if spatial_algorithm in ["ISS", "ISS1", "ISS2"]:
            assert ilrma.demix_filter is None

        print(ilrma)



import os
import sys
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pytest
import scipy.signal as ss

from ssspy.bss.iva import (
    PDSIVA,
    AuxGaussIVA,
    AuxIVA,
    AuxIVABase,
    AuxLaplaceIVA,
    FasterIVA,
    FastIVA,
    FastIVABase,
    GradGaussIVA,
    GradIVA,
    GradIVABase,
    GradLaplaceIVA,
    IVABase,
    NaturalGradGaussIVA,
    NaturalGradIVA,
    NaturalGradLaplaceIVA,
)

ssspy_tests_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
sys.path.append(ssspy_tests_dir)

from dummy.callback import DummyCallback, dummy_function  # noqa: E402
from dummy.utils.dataset import download_sample_speech_data  # noqa: E402

max_duration = 0.5
n_fft = 512
hop_length = 256
n_bins = n_fft // 2 + 1
n_iter = 3

parameters_spatial_algorithm = ["IP", "IP1", "IP2", "ISS", "ISS1", "ISS2", "IPA"]
parameters_callbacks = [None, dummy_function, [DummyCallback(), dummy_function]]
parameters_is_holonomic = [True, False]
parameters_scale_restoration = [True, False, "projection_back", "minimal_distortion_principle"]
parameters_grad_iva = [
    (2, {}),
    (
        3,
        {"demix_filter": np.tile(-np.eye(3, dtype=np.complex128), reps=(n_bins, 1, 1))},
    ),
]
parameters_fast_iva = [
    (2, "dev1_female3", {}),
    (
        3,
        "dev1_female3",
        {"demix_filter": np.tile(-np.eye(3, dtype=np.complex128), reps=(n_bins, 1, 1))},
    ),
    (2, "dev1_female3", {"demix_filter": None}),
]
parameters_aux_iva = [
    (2, "dev1_female3", {}),
    (
        3,
        "dev1_female3",
        {"demix_filter": np.tile(-np.eye(3, dtype=np.complex128), reps=(n_bins, 1, 1))},
    ),
    (2, "dev1_female3", {"demix_filter": None}),
    (
        3,
        "dev1_female3",
        {"demix_filter": np.tile(-np.eye(3, dtype=np.complex128), reps=(n_bins, 1, 1))},
    ),
    (4, "dev1_female4", {"demix_filter": None}),
]
parameters_pds_iva = [
    (2, "dev1_female3", {}),
    (
        3,
        "dev1_female3",
        {"demix_filter": np.tile(-np.eye(3, dtype=np.complex128), reps=(n_bins, 1, 1))},
    ),
    (4, "dev1_female4", {}),
]


@pytest.mark.parametrize("callbacks", parameters_callbacks)
def test_iva_base(
    callbacks: Optional[Union[Callable[[AuxIVA], None], List[Callable[[AuxIVA], None]]]],
):
    iva = IVABase(callbacks=callbacks)

    print(iva)


@pytest.mark.parametrize("callbacks", parameters_callbacks)
def test_fast_iva_base(
    callbacks: Optional[Union[Callable[[AuxIVA], None], List[Callable[[AuxIVA], None]]]],
):
    np.random.seed(111)

    iva = FastIVABase(callbacks=callbacks)

    print(iva)


@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_grad_iva_base(
    callbacks: Optional[Union[Callable[[GradIVA], None], List[Callable[[GradIVA], None]]]],
    is_holonomic: bool,
):
    np.random.seed(111)

    def contrast_fn(y: np.ndarray) -> np.ndarray:
        r"""Contrast function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_bins, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_frames).
        """
        return 2 * np.linalg.norm(y, axis=1)

    def score_fn(y) -> np.ndarray:
        r"""Score function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_bins, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_bins, n_frames).
        """
        norm = np.linalg.norm(y, axis=1, keepdims=True)
        norm = np.maximum(norm, 1e-10)
        return y / norm

    iva = GradIVABase(
        contrast_fn=contrast_fn, score_fn=score_fn, callbacks=callbacks, is_holonomic=is_holonomic
    )

    print(iva)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_iva)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_grad_iva(
    n_sources: int,
    callbacks: Optional[Union[Callable[[GradIVA], None], List[Callable[[GradIVA], None]]]],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    def contrast_fn(y: np.ndarray) -> np.ndarray:
        r"""Contrast function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_bins, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_frames).
        """
        return 2 * np.linalg.norm(y, axis=1)

    def score_fn(y) -> np.ndarray:
        r"""Score function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_bins, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_bins, n_frames).
        """
        norm = np.linalg.norm(y, axis=1, keepdims=True)
        norm = np.maximum(norm, 1e-10)
        return y / norm

    iva = GradIVA(
        contrast_fn=contrast_fn, score_fn=score_fn, callbacks=callbacks, is_holonomic=is_holonomic
    )
    spectrogram_est = iva(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(iva.loss[-1]) is float

    print(iva)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_iva)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_natural_grad_iva(
    n_sources: int,
    callbacks: Optional[
        Union[Callable[[NaturalGradIVA], None], List[Callable[[NaturalGradIVA], None]]]
    ],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    def contrast_fn(y: np.ndarray) -> np.ndarray:
        r"""Contrast function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_bins, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_frames).
        """
        return 2 * np.linalg.norm(y, axis=1)

    def score_fn(y):
        r"""Score function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_bins, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_bins, n_frames).
        """
        norm = np.linalg.norm(y, axis=1, keepdims=True)
        norm = np.maximum(norm, 1e-10)
        return y / norm

    iva = NaturalGradIVA(
        contrast_fn=contrast_fn, score_fn=score_fn, callbacks=callbacks, is_holonomic=is_holonomic
    )
    spectrogram_est = iva(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(iva.loss[-1]) is float

    print(iva)


@pytest.mark.parametrize("n_sources, sisec2010_tag, reset_kwargs", parameters_fast_iva)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
def test_fast_iva(
    n_sources: int,
    sisec2010_tag: str,
    callbacks: Optional[Union[Callable[[AuxIVA], None], List[Callable[[AuxIVA], None]]]],
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    def contrast_fn(y: np.ndarray) -> np.ndarray:
        r"""Contrast function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_bins, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_frames).
        """
        return 2 * np.linalg.norm(y, axis=1)

    def d_contrast_fn(y) -> np.ndarray:
        r"""Derivative of contrast function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_frames).
        """
        return 2 * np.ones_like(y)

    def dd_contrast_fn(y) -> np.ndarray:
        r"""Second order derivative of contrast function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_frames).
        """
        return 2 * np.zeros_like(y)

    iva = FastIVA(
        contrast_fn=contrast_fn,
        d_contrast_fn=d_contrast_fn,
        dd_contrast_fn=dd_contrast_fn,
        callbacks=callbacks,
    )
    spectrogram_est = iva(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(iva.loss[-1]) is float

    print(iva)


@pytest.mark.parametrize("n_sources, sisec2010_tag, reset_kwargs", parameters_fast_iva)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
def test_faster_iva(
    n_sources: int,
    sisec2010_tag: str,
    callbacks: Optional[Union[Callable[[AuxIVA], None], List[Callable[[AuxIVA], None]]]],
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    def contrast_fn(y: np.ndarray) -> np.ndarray:
        r"""Contrast function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_bins, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_frames).
        """
        return 2 * np.linalg.norm(y, axis=1)

    def d_contrast_fn(y) -> np.ndarray:
        r"""Derivative of contrast function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_frames).
        """
        return 2 * np.ones_like(y)

    iva = FasterIVA(contrast_fn=contrast_fn, d_contrast_fn=d_contrast_fn, callbacks=callbacks)
    spectrogram_est = iva(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(iva.loss[-1]) is float

    print(iva)


@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_aux_iva_base(
    callbacks: Optional[Union[Callable[[AuxIVA], None], List[Callable[[AuxIVA], None]]]],
    scale_restoration: Union[str, bool],
):
    np.random.seed(111)

    def contrast_fn(y: np.ndarray) -> np.ndarray:
        r"""Contrast function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_bins, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_frames).
        """
        return 2 * np.linalg.norm(y, axis=1)

    def d_contrast_fn(y) -> np.ndarray:
        r"""Derivative of contrast function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_frames).
        """
        return 2 * np.ones_like(y)

    iva = AuxIVABase(
        contrast_fn=contrast_fn,
        d_contrast_fn=d_contrast_fn,
        callbacks=callbacks,
        scale_restoration=scale_restoration,
    )

    print(iva)


@pytest.mark.parametrize("n_sources, sisec2010_tag, reset_kwargs", parameters_aux_iva)
@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_aux_iva(
    n_sources: int,
    sisec2010_tag: str,
    spatial_algorithm: str,
    callbacks: Optional[Union[Callable[[AuxIVA], None], List[Callable[[AuxIVA], None]]]],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[Any, Any],
):
    if spatial_algorithm in ["IP", "ISS"] and not pytest.run_redundant:
        pytest.skip(reason="Need --run-redundant option to run.")

    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    def contrast_fn(y: np.ndarray) -> np.ndarray:
        r"""Contrast function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_bins, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_frames).
        """
        return 2 * np.linalg.norm(y, axis=1)

    def d_contrast_fn(y) -> np.ndarray:
        r"""Derivative of contrast function.

        Args:
            y (np.ndarray):
                The shape is (n_sources, n_frames).

        Returns:
            np.ndarray:
                The shape is (n_sources, n_frames).
        """
        return 2 * np.ones_like(y)

    iva = AuxIVA(
        spatial_algorithm=spatial_algorithm,
        contrast_fn=contrast_fn,
        d_contrast_fn=d_contrast_fn,
        callbacks=callbacks,
        scale_restoration=scale_restoration,
    )
    spectrogram_est = iva(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(iva.loss[-1]) is float

    if spatial_algorithm in ["ISS", "ISS1", "ISS2"]:
        assert iva.demix_filter is None

    print(iva)


@pytest.mark.parametrize("n_sources, sisec2010_tag, reset_kwargs", parameters_pds_iva)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_pds_iva(
    n_sources: int,
    sisec2010_tag: str,
    callbacks: Optional[Union[Callable[[AuxIVA], None], List[Callable[[AuxIVA], None]]]],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    iva = PDSIVA(
        contrast_fn=None,
        prox_penalty=None,
        callbacks=callbacks,
        scale_restoration=scale_restoration,
    )
    spectrogram_est = iva(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(iva.loss[-1]) is float

    print(iva)


@pytest.mark.parametrize("specify_contrast_fn", [True, False])
def test_iva_insufficient_fn(specify_contrast_fn: bool):
    def _contrast_fn(y: np.ndarray) -> np.ndarray:
        return np.linalg.norm(y, axis=1)

    def _prox_penalty(y: np.ndarray, step_size: float = 1) -> np.ndarray:
        norm = np.linalg.norm(y, axis=1, keepdims=True)
        return y * np.maximum(1 - step_size / norm, 0)

    if specify_contrast_fn:
        contrast_fn = _contrast_fn
        prox_penalty = None
    else:
        contrast_fn = None
        prox_penalty = _prox_penalty

    with pytest.raises(ValueError) as e:
        _ = PDSIVA(
            contrast_fn=contrast_fn,
            prox_penalty=prox_penalty,
        )

    if specify_contrast_fn:
        assert str(e.value) == "Set prox_penalty."
    else:
        assert str(e.value) == "Set contrast_fn."


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_iva)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_grad_laplace_iva(
    n_sources: int,
    callbacks: Optional[
        Union[Callable[[GradLaplaceIVA], None], List[Callable[[GradLaplaceIVA], None]]]
    ],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    iva = GradLaplaceIVA(callbacks=callbacks, is_holonomic=is_holonomic)
    spectrogram_est = iva(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(iva.loss[-1]) is float

    print(iva)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_iva)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_grad_gauss_iva(
    n_sources: int,
    callbacks: Optional[
        Union[Callable[[GradGaussIVA], None], List[Callable[[GradGaussIVA], None]]]
    ],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    iva = GradGaussIVA(callbacks=callbacks, is_holonomic=is_holonomic)
    spectrogram_est = iva(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(iva.loss[-1]) is float

    print(iva)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_iva)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_natural_grad_laplace_iva(
    n_sources: int,
    callbacks: Optional[
        Union[
            Callable[[NaturalGradLaplaceIVA], None], List[Callable[[NaturalGradLaplaceIVA], None]]
        ]
    ],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    iva = NaturalGradLaplaceIVA(callbacks=callbacks, is_holonomic=is_holonomic)
    spectrogram_est = iva(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(iva.loss[-1]) is float

    print(iva)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_iva)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_natural_grad_gauss_iva(
    n_sources: int,
    callbacks: Optional[
        Union[Callable[[NaturalGradGaussIVA], None], List[Callable[[NaturalGradGaussIVA], None]]]
    ],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    iva = NaturalGradGaussIVA(callbacks=callbacks, is_holonomic=is_holonomic)
    spectrogram_est = iva(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(iva.loss[-1]) is float

    print(iva)


@pytest.mark.parametrize("n_sources, sisec2010_tag, reset_kwargs", parameters_aux_iva)
@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_aux_laplace_iva(
    n_sources: int,
    sisec2010_tag: str,
    spatial_algorithm: str,
    callbacks: Optional[
        Union[
            Callable[[NaturalGradLaplaceIVA], None], List[Callable[[NaturalGradLaplaceIVA], None]]
        ]
    ],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[Any, Any],
):
    if spatial_algorithm in ["IP", "ISS"] and not pytest.run_redundant:
        pytest.skip(reason="Need --run-redundant option to run.")

    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    iva = AuxLaplaceIVA(
        spatial_algorithm=spatial_algorithm,
        callbacks=callbacks,
        scale_restoration=scale_restoration,
    )
    spectrogram_est = iva(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(iva.loss[-1]) is float

    if spatial_algorithm in ["ISS", "ISS1", "ISS2"]:
        assert iva.demix_filter is None

    print(iva)


@pytest.mark.parametrize("n_sources, sisec2010_tag, reset_kwargs", parameters_aux_iva)
@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_aux_gauss_iva(
    n_sources: int,
    sisec2010_tag: str,
    spatial_algorithm: str,
    callbacks: Optional[
        Union[Callable[[NaturalGradGaussIVA], None], List[Callable[[NaturalGradGaussIVA], None]]]
    ],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[Any, Any],
):
    if spatial_algorithm in ["IP", "ISS"] and not pytest.run_redundant:
        pytest.skip(reason="Need --run-redundant option to run.")

    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    iva = AuxGaussIVA(
        spatial_algorithm=spatial_algorithm,
        callbacks=callbacks,
        scale_restoration=scale_restoration,
    )
    spectrogram_est = iva(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(iva.loss[-1]) is float

    if spatial_algorithm in ["ISS", "ISS1", "ISS2"]:
        assert iva.demix_filter is None

    print(iva)



import os
import sys
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pytest
import scipy.signal as ss

from ssspy.bss.ipsdta import TIPSDTA, BlockDecompositionIPSDTABase, GaussIPSDTA, IPSDTABase

ssspy_tests_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
sys.path.append(ssspy_tests_dir)

from dummy.callback import DummyCallback, dummy_function  # noqa: E402
from dummy.utils.dataset import download_sample_speech_data  # noqa: E402

max_duration = 0.1
n_fft = 256
hop_length = 128
window = "hann"
n_bins = n_fft // 2 + 1
n_iter = 3
rng = np.random.default_rng(42)

parameters_dof = [100]
parameters_spatial_algorithm = ["FPI", "VCD"]
parameters_callbacks = [None, dummy_function, [DummyCallback(), dummy_function]]
parameters_source_normalization = [True, False]
parameters_scale_restoration = [True, False, "projection_back", "minimal_distortion_principle"]
parameters_ipsdta_base = [2]
parameters_block_decomposition_ipsdta_base = [4]
parameters_ipsdta = [
    (
        2,
        2,
        43,
        {
            "demix_filter": np.tile(np.eye(2, dtype=np.complex128), (n_bins, 1, 1)),
        },
    ),
    (3, 2, 64, {}),
]


@pytest.mark.parametrize(
    "n_basis",
    parameters_ipsdta_base,
)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_ipsdta_base(
    n_basis: int,
    callbacks: Optional[Union[Callable[[IPSDTABase], None], List[Callable[[IPSDTABase], None]]]],
    scale_restoration: Union[str, bool],
):
    ipsdta = IPSDTABase(
        n_basis,
        callbacks=callbacks,
        scale_restoration=scale_restoration,
        record_loss=False,
        rng=rng,
    )

    print(ipsdta)


@pytest.mark.parametrize(
    "n_basis",
    parameters_ipsdta_base,
)
@pytest.mark.parametrize(
    "n_blocks",
    parameters_block_decomposition_ipsdta_base,
)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_block_decomposition_ipsdta_base(
    n_basis: int,
    n_blocks: int,
    callbacks: Optional[
        Union[
            Callable[[BlockDecompositionIPSDTABase], None],
            List[Callable[[BlockDecompositionIPSDTABase], None]],
        ]
    ],
    scale_restoration: Union[str, bool],
):
    ipsdta = BlockDecompositionIPSDTABase(
        n_basis,
        n_blocks,
        callbacks=callbacks,
        scale_restoration=scale_restoration,
        record_loss=False,
        rng=rng,
    )

    print(ipsdta)


@pytest.mark.parametrize(
    "n_sources, n_basis, n_blocks, reset_kwargs",
    parameters_ipsdta,
)
@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("source_normalization", parameters_source_normalization)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_gauss_ipsdta(
    n_sources: int,
    n_basis: int,
    n_blocks: int,
    spatial_algorithm: str,
    callbacks: Optional[Union[Callable[[GaussIPSDTA], None], List[Callable[[GaussIPSDTA], None]]]],
    source_normalization: Optional[Union[str, bool]],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[str, Any],
):
    if n_sources < 4:
        sisec2010_tag = "dev1_female3"
    elif n_sources == 4:
        sisec2010_tag = "dev1_female4"
    else:
        raise ValueError("n_sources should be less than 5.")

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window=window, nperseg=n_fft, noverlap=n_fft - hop_length
    )

    ipsdta = GaussIPSDTA(
        n_basis,
        n_blocks,
        spatial_algorithm=spatial_algorithm,
        callbacks=callbacks,
        source_normalization=source_normalization,
        scale_restoration=scale_restoration,
        rng=rng,
    )

    if spatial_algorithm == "FPI":
        with pytest.raises(NotImplementedError) as e:
            spectrogram_est = ipsdta(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

        assert str(e.value) == "IPSDTA with fixed-point iteration is not supported."
    else:
        spectrogram_est = ipsdta(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

        assert spectrogram_mix.shape == spectrogram_est.shape
        assert type(ipsdta.loss[-1]) is float

    print(ipsdta)


@pytest.mark.parametrize(
    "n_sources, n_basis, n_blocks, reset_kwargs",
    parameters_ipsdta,
)
@pytest.mark.parametrize("dof", parameters_dof)
@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("source_normalization", parameters_source_normalization)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_t_ipsdta(
    n_sources: int,
    n_basis: int,
    n_blocks: int,
    dof: float,
    spatial_algorithm: str,
    callbacks: Optional[Union[Callable[[GaussIPSDTA], None], List[Callable[[GaussIPSDTA], None]]]],
    source_normalization: Optional[Union[str, bool]],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[str, Any],
):
    if n_sources < 4:
        sisec2010_tag = "dev1_female3"
    elif n_sources == 4:
        sisec2010_tag = "dev1_female4"
    else:
        raise ValueError("n_sources should be less than 5.")

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window=window, nperseg=n_fft, noverlap=n_fft - hop_length
    )

    ipsdta = TIPSDTA(
        n_basis,
        n_blocks,
        dof=dof,
        spatial_algorithm=spatial_algorithm,
        callbacks=callbacks,
        source_normalization=source_normalization,
        scale_restoration=scale_restoration,
        rng=rng,
    )

    if spatial_algorithm != "VCD":
        with pytest.raises(NotImplementedError) as e:
            spectrogram_est = ipsdta(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

        assert str(e.value) == "Not support {}.".format(spatial_algorithm)
    else:
        spectrogram_est = ipsdta(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

        assert spectrogram_mix.shape == spectrogram_est.shape
        assert type(ipsdta.loss[-1]) is float

    print(ipsdta)



import numpy as np

from ssspy.linalg.polynomial import _find_cubic_roots, solve_cubic


def test_find_cubic_roots():
    rng = np.random.default_rng(0)

    n_bins, n_channels = 3, 2

    P = rng.standard_normal((n_bins, n_channels))
    Q = rng.standard_normal((n_bins, n_channels))
    X = _find_cubic_roots(P, Q)
    Y = X**3 + P * X + Q

    assert np.allclose(Y, 0)


def test_solve_cubic():
    rng = np.random.default_rng(0)

    n_bins, n_channels = 3, 2

    # real coefficients
    A = rng.standard_normal((n_bins, n_channels))
    B = rng.standard_normal((n_bins, n_channels))
    C = rng.standard_normal((n_bins, n_channels))
    D = rng.standard_normal((n_bins, n_channels))

    X = solve_cubic(A, B, C)
    Y = X**3 + A * X**2 + B * X + C

    assert np.allclose(Y, 0)

    X = solve_cubic(A, B, C, D)
    Y = A * X**3 + B * X**2 + C * X + D

    assert np.allclose(Y, 0)

    # corner case
    A = np.zeros_like(C)
    B = np.zeros_like(C)

    X = solve_cubic(A, B, C)
    Y = X**3 + A * X**2 + B * X + C

    assert np.allclose(Y, 0)

    # complex coefficients
    A = rng.standard_normal((n_bins, n_channels)) + 1j * rng.standard_normal((n_bins, n_channels))
    B = rng.standard_normal((n_bins, n_channels)) + 1j * rng.standard_normal((n_bins, n_channels))
    C = rng.standard_normal((n_bins, n_channels)) + 1j * rng.standard_normal((n_bins, n_channels))
    D = rng.standard_normal((n_bins, n_channels)) + 1j * rng.standard_normal((n_bins, n_channels))

    X = solve_cubic(A, B, C)
    Y = X**3 + A * X**2 + B * X + C

    assert np.allclose(Y, 0)

    X = solve_cubic(A, B, C, D)
    Y = A * X**3 + B * X**2 + C * X + D

    assert np.allclose(Y, 0)

    # corner case
    A = np.zeros_like(C)
    B = np.zeros_like(C)

    X = solve_cubic(A, B, C)
    Y = X**3 + A * X**2 + B * X + C

    assert np.allclose(Y, 0)



from typing import Callable, Optional

import numpy as np

from .eigh import eigh


def sqrtmh(X: np.ndarray) -> np.ndarray:
    r"""Compute square root of a positive semidefinite Hermitian or symmetric matrix.

    Args:
        X (numpy.ndarray):
            A complex Hermitian or symmetric matrix with shape of (\*, n_channels, n_channels).

    Returns:
        numpy.ndarray of square root. The shape is same as that of input.
    """
    Lamb, P = eigh(X)

    P_Hermite = P.swapaxes(-2, -1)

    if np.iscomplexobj(X):
        P_Hermite = P_Hermite.conj()

    Lamb = np.sqrt(Lamb)[..., np.newaxis] * np.eye(Lamb.shape[-1])

    return P @ Lamb @ P_Hermite


def invsqrtmh(
    X: np.ndarray,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = None,
) -> np.ndarray:
    r"""Compute inversion of square root for a positive definite Hermitian or symmetric matrix.

    Args:
        X (numpy.ndarray):
            A complex Hermitian matrix with shape of (\*, n_channels, n_channels).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to receive and return the same shape as that of X.
            By default, the identity function (``lambda x: x``) is used.

    Returns:
        numpy.ndarray of inversion of square root. The shape is same as that of input.
    """

    def _identity(x):
        return x

    if flooring_fn is None:
        flooring_fn = _identity

    Lamb, P = eigh(X)

    P_Hermite = P.swapaxes(-2, -1)

    if np.iscomplexobj(X):
        P_Hermite = P_Hermite.conj()

    Lamb = 1 / flooring_fn(np.sqrt(Lamb))
    Lamb = Lamb[..., np.newaxis] * np.eye(Lamb.shape[-1])

    return P @ Lamb @ P_Hermite



import numpy as np

from ssspy.bss._update_spatial_model import update_by_ipa


def main() -> None:
    n_sources, n_bins, n_frames = 3, 1, 100
    rng = np.random.default_rng(0)
    separated = rng.standard_normal((n_sources, n_bins, n_frames)) + 1j * rng.standard_normal(
        (n_sources, n_bins, n_frames)
    )  # (n_sources, n_bins, n_frames)
    weight = rng.uniform(0, 1, (n_sources, 1, n_frames))

    separated = update_by_ipa(separated, weight, flooring_fn=None)


if __name__ == "__main__":
    main()



import numpy as np
import pytest

from ssspy.linalg import eigh, eigh2

parameters_sources = [2, 5]
parameters_channels = [4, 3]
parameters_frames = [32, 16]
parameters_is_complex = [True, False]
parameters_type = [1, 2, 3]


@pytest.mark.parametrize("n_sources", parameters_sources)
@pytest.mark.parametrize("n_channels", parameters_channels)
@pytest.mark.parametrize("n_frames", parameters_frames)
@pytest.mark.parametrize("is_complex", parameters_is_complex)
def test_eigh(n_sources: int, n_channels: int, n_frames: int, is_complex: bool):
    np.random.seed(111)

    shape = (n_sources, n_channels, n_frames)

    if is_complex:
        a = np.random.randn(*shape) + 1j * np.random.randn(*shape)
        A = np.mean(a[:, :, np.newaxis, :] * a[:, np.newaxis, :, :].conj(), axis=-1)
    else:
        a = np.random.randn(*shape)
        A = np.mean(a[:, :, np.newaxis, :] * a[:, np.newaxis, :, :], axis=-1)

    lamb, z = eigh(A)

    assert lamb.shape == (n_sources, n_channels)
    assert z.shape == (n_sources, n_channels, n_channels)

    assert np.allclose(A @ z, lamb[:, np.newaxis, :] * z)


@pytest.mark.parametrize("n_sources", parameters_sources)
@pytest.mark.parametrize("n_channels", parameters_channels)
@pytest.mark.parametrize("n_frames", parameters_frames)
@pytest.mark.parametrize("is_complex", parameters_is_complex)
@pytest.mark.parametrize("type", parameters_type)
def test_generalized_eigh(
    n_sources: int, n_channels: int, n_frames: int, is_complex: bool, type: int
):
    np.random.seed(111)

    shape = (n_sources, n_channels, n_frames)

    if is_complex:
        a = np.random.randn(*shape) + 1j * np.random.randn(*shape)
        b = np.random.randn(*shape) + 1j * np.random.randn(*shape)
        A = np.mean(a[:, :, np.newaxis, :] * a[:, np.newaxis, :, :].conj(), axis=-1)
        B = np.mean(b[:, :, np.newaxis, :] * b[:, np.newaxis, :, :].conj(), axis=-1)
    else:
        a = np.random.randn(*shape)
        b = np.random.randn(*shape)
        A = np.mean(a[:, :, np.newaxis, :] * a[:, np.newaxis, :, :], axis=-1)
        B = np.mean(b[:, :, np.newaxis, :] * b[:, np.newaxis, :, :], axis=-1)

    lamb, z = eigh(A, B, type=type)

    assert lamb.shape == (n_sources, n_channels)
    assert z.shape == (n_sources, n_channels, n_channels)

    if type == 1:
        assert np.allclose(A @ z, lamb[:, np.newaxis, :] * (B @ z))
    elif type == 2:
        assert np.allclose(A @ B @ z, lamb[:, np.newaxis, :] * z)
    elif type == 3:
        assert np.allclose(B @ A @ z, lamb[:, np.newaxis, :] * z)
    else:
        raise ValueError("Invalid type={} is given.".format(type))


@pytest.mark.parametrize("n_sources", parameters_sources)
@pytest.mark.parametrize("n_frames", parameters_frames)
@pytest.mark.parametrize("is_complex", parameters_is_complex)
def test_eigh2(n_sources: int, n_frames: int, is_complex: bool):
    np.random.seed(111)

    shape = (n_sources, 2, n_frames)

    if is_complex:
        a = np.random.randn(*shape) + 1j * np.random.randn(*shape)
        A = np.mean(a[:, :, np.newaxis, :] * a[:, np.newaxis, :, :].conj(), axis=-1)
    else:
        a = np.random.randn(*shape)
        A = np.mean(a[:, :, np.newaxis, :] * a[:, np.newaxis, :, :], axis=-1)

    lamb, z = eigh2(A)

    assert lamb.shape == (n_sources, 2)
    assert z.shape == (n_sources, 2, 2)

    assert np.allclose(A @ z, lamb[:, np.newaxis, :] * z)


@pytest.mark.parametrize("n_sources", parameters_sources)
@pytest.mark.parametrize("n_frames", parameters_frames)
@pytest.mark.parametrize("is_complex", parameters_is_complex)
@pytest.mark.parametrize("type", parameters_type)
def test_generalized_eigh2(n_sources: int, n_frames: int, is_complex: bool, type: int):
    np.random.seed(111)

    shape = (n_sources, 2, n_frames)

    if is_complex:
        a = np.random.randn(*shape) + 1j * np.random.randn(*shape)
        b = np.random.randn(*shape) + 1j * np.random.randn(*shape)
        A = np.mean(a[:, :, np.newaxis, :] * a[:, np.newaxis, :, :].conj(), axis=-1)
        B = np.mean(b[:, :, np.newaxis, :] * b[:, np.newaxis, :, :].conj(), axis=-1)
    else:
        a = np.random.randn(*shape)
        b = np.random.randn(*shape)
        A = np.mean(a[:, :, np.newaxis, :] * a[:, np.newaxis, :, :], axis=-1)
        B = np.mean(b[:, :, np.newaxis, :] * b[:, np.newaxis, :, :], axis=-1)

    lamb, z = eigh2(A, B, type=type)

    assert lamb.shape == (n_sources, 2)
    assert z.shape == (n_sources, 2, 2)

    if type == 1:
        assert np.allclose(A @ z, lamb[:, np.newaxis, :] * (B @ z))
    elif type == 2:
        assert np.allclose(A @ B @ z, lamb[:, np.newaxis, :] * z)
    elif type == 3:
        assert np.allclose(B @ A @ z, lamb[:, np.newaxis, :] * z)
    else:
        raise ValueError("Invalid type={} is given.".format(type))



import numpy as np

from ssspy.linalg.lqpqm import _find_largest_root


def test_find_largest_root():
    alpha = np.array([-1, 1, 1, -1 + 1j])
    beta = np.array([0, 1, 1, -1 - 1j])
    gamma = np.array([1, 1, 2, 1])

    A = -np.real(alpha + beta + gamma)
    B = np.real(alpha * beta + beta * gamma + gamma * alpha)
    C = -np.real(alpha * beta * gamma)

    X = _find_largest_root(A, B, C)

    assert np.allclose(X, gamma)



import hashlib
import os
from typing import Tuple

import numpy as np

from ssspy.utils.dataset import download_sample_speech_data as _download


def download_sample_speech_data(
    sisec2010_root: str = "./tests/.data/SiSEC2010",
    mird_root: str = "./tests/.data/MIRD",
    n_sources: int = 3,
    sisec2010_tag: str = "dev1_female3",
    max_duration: float = 10,
    conv: bool = True,
    cache_dir: str = ".tests/.data/.cache",
) -> Tuple[np.ndarray, int]:
    hash = hashlib.sha256(sisec2010_root.encode("utf-8")).hexdigest()
    hash += hashlib.sha256(mird_root.encode("utf-8")).hexdigest()
    hash += hashlib.sha256(str(n_sources).encode("utf-8")).hexdigest()
    hash += hashlib.sha256(sisec2010_tag.encode("utf-8")).hexdigest()
    hash += hashlib.sha256(str(max_duration).encode("utf-8")).hexdigest()
    hash += hashlib.sha256(str(conv).encode("utf-8")).hexdigest()

    # because concatenated hash is too long
    hash = hashlib.sha256(hash.encode("utf-8")).hexdigest()

    npz_path = os.path.join(cache_dir, "{}.npz".format(hash))

    if os.path.exists(npz_path):
        npz = np.load(npz_path)
        waveform_src_img, sample_rate = npz["waveform_src_img"], npz["sample_rate"]
        sample_rate = sample_rate.item()
    else:
        waveform_src_img, sample_rate = _download(
            sisec2010_root=sisec2010_root,
            mird_root=mird_root,
            n_sources=n_sources,
            sisec2010_tag=sisec2010_tag,
            max_duration=max_duration,
            conv=conv,
        )
        os.makedirs(cache_dir, exist_ok=True)
        np.savez(npz_path, waveform_src_img=waveform_src_img, sample_rate=sample_rate)

    return waveform_src_img, sample_rate



import numpy as np
import pytest

from ssspy.transform import pca

parameters_ascend = [True, False]
parameters_batch_size = [1, 4]
parameters_n_channels = [2, 3]
parameters_pca_real = [10, 20]
parameters_pca_complex = [(257, 8), (65, 12)]


@pytest.mark.parametrize("ascend", parameters_ascend)
@pytest.mark.parametrize("n_channels", parameters_n_channels)
@pytest.mark.parametrize("n_samples", parameters_pca_real)
def test_pca_real_2d(ascend: bool, n_channels: int, n_samples: int):
    np.random.seed(111)

    input = np.random.randn(n_channels, n_samples)
    output = pca(input, ascend=ascend)

    assert input.shape == output.shape

    covariance = output[:, np.newaxis, :] * output[np.newaxis, :, :]
    covariance = np.mean(covariance, axis=-1)
    mask = 1 - np.eye(n_channels)
    zero = np.zeros((n_channels, n_channels))

    assert np.allclose(mask * covariance, zero)


@pytest.mark.parametrize("ascend", parameters_ascend)
@pytest.mark.parametrize("batch_size", parameters_batch_size)
@pytest.mark.parametrize("n_channels", parameters_n_channels)
@pytest.mark.parametrize("n_samples", parameters_pca_real)
def test_pca_real_3d(ascend: bool, batch_size: int, n_channels: int, n_samples: int):
    np.random.seed(111)

    input = np.random.randn(batch_size, n_channels, n_samples)
    output = pca(input, ascend=ascend)

    assert input.shape == output.shape

    covariance = output[:, :, np.newaxis, :] * output[:, np.newaxis, :, :]
    covariance = np.mean(covariance, axis=-1)
    mask = 1 - np.eye(n_channels)
    zero = np.zeros((batch_size, n_channels, n_channels))

    assert np.allclose(mask * covariance, zero)


@pytest.mark.parametrize("ascend", parameters_ascend)
@pytest.mark.parametrize("n_channels", parameters_n_channels)
@pytest.mark.parametrize("n_bins, n_frames", parameters_pca_complex)
def test_pca_complex_3d(ascend: bool, n_channels: int, n_bins: int, n_frames: int):
    np.random.seed(111)

    real = np.random.randn(n_channels, n_bins, n_frames)
    imag = np.random.randn(n_channels, n_bins, n_frames)
    input = real + 1j * imag
    output = pca(input, ascend=ascend)

    assert input.shape == output.shape

    covariance = output[:, np.newaxis, :, :] * output[np.newaxis, :, :, :].conj()
    covariance = np.mean(covariance, axis=-1)
    covariance = covariance.transpose(2, 0, 1)
    mask = 1 - np.eye(n_channels)
    zero = np.zeros((n_bins, n_channels, n_channels))

    assert np.allclose(mask * covariance, zero)


@pytest.mark.parametrize("ascend", parameters_ascend)
@pytest.mark.parametrize("batch_size", parameters_batch_size)
@pytest.mark.parametrize("n_channels", parameters_n_channels)
@pytest.mark.parametrize("n_bins, n_frames", parameters_pca_complex)
def test_pca_complex_4d(ascend: bool, batch_size: int, n_channels: int, n_bins: int, n_frames: int):
    np.random.seed(111)

    real = np.random.randn(batch_size, n_channels, n_bins, n_frames)
    imag = np.random.randn(batch_size, n_channels, n_bins, n_frames)
    input = real + 1j * imag
    output = pca(input, ascend=ascend)

    assert input.shape == output.shape

    covariance = output[:, :, np.newaxis, :, :] * output[:, np.newaxis, :, :, :].conj()
    covariance = np.mean(covariance, axis=-1)
    covariance = covariance.transpose(0, 3, 1, 2)
    mask = 1 - np.eye(n_channels)
    zero = np.zeros((batch_size, n_bins, n_channels, n_channels))

    assert np.allclose(mask * covariance, zero)



import functools
import warnings
from typing import Callable, Optional

import numpy as np

from ..algorithm.permutation_alignment import (
    correlation_based_permutation_solver as correlation_based_permutation_solver_base,
)
from ..special.flooring import max_flooring

EPS = 1e-10


def correlation_based_permutation_solver(
    separated: np.ndarray,
    *args,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
    overwrite: bool = True,
) -> np.ndarray:
    r"""Solve permutaion of estimated spectrograms."""

    warnings.warn(
        "Use ssspy.algorithm.permutation_alignment.correlation_based_permutation_solver instead.",
        UserWarning,
    )

    return correlation_based_permutation_solver_base(
        separated, *args, flooring_fn=flooring_fn, overwrite=overwrite
    )



from typing import Optional

import numpy as np
import pytest

from ssspy.algorithm import minimal_distortion_principle

parameters = [(2, 0), (3, 2), (2, None)]


@pytest.mark.parametrize("n_sources, reference_id", parameters)
def test_minimal_distortion_principle(n_sources: int, reference_id: Optional[int]):
    rng = np.random.default_rng(0)

    n_channels = n_sources
    n_bins, n_frames = 5, 8

    spectrogram_mix = rng.standard_normal(
        (n_channels, n_bins, n_frames)
    ) + 1j * rng.standard_normal((n_channels, n_bins, n_frames))
    demix_filter = rng.standard_normal((n_bins, n_sources, n_channels)) + 1j * rng.standard_normal(
        (n_bins, n_sources, n_channels)
    )
    spectrogram_est = demix_filter @ spectrogram_mix.transpose(1, 0, 2)
    spectrogram_est = spectrogram_est.transpose(1, 0, 2)

    spectrogram_est_scaled = minimal_distortion_principle(
        spectrogram_est, spectrogram_mix, reference_id=reference_id
    )

    if reference_id is None:
        for _spectrogram_est_scaled in spectrogram_est_scaled:
            assert spectrogram_mix.shape == _spectrogram_est_scaled.shape
    else:
        assert spectrogram_mix.shape == spectrogram_est.shape



from typing import Callable, List, Optional, Union

import numpy as np

from ..algorithm import (
    MINIMAL_DISTORTION_PRINCIPLE_KEYWORDS,
    PROJECTION_BACK_KEYWORDS,
    minimal_distortion_principle,
    projection_back,
)
from ..linalg import prox

EPS = 1e-10


class PDSBSSBase:
    r"""Base class of blind source separation \
    via proximal splitting algorithm [#yatabe2018determined]_.

    Args:
        penalty_fn (callable):
            Penalty function that determines source model.
        prox_penalty (callable):
            Proximal operator of penalty function.
            Default: ``None``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back`` explicitly.
            Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back.
            Default: ``0``.

    .. [#yatabe2018determined] K. Yatabe and D. Kitamura,
        "Determined blind source separation via proximal splitting algorithm,"
        in *Proc of ICASSP*, pp. 776-780, 2018.
    """

    def __init__(
        self,
        penalty_fn: Callable[[np.ndarray, np.ndarray], float] = None,
        prox_penalty: Callable[[np.ndarray, float], np.ndarray] = None,
        callbacks: Optional[
            Union[Callable[["PDSBSSBase"], None], List[Callable[["PDSBSSBase"], None]]]
        ] = None,
        scale_restoration: bool = True,
        record_loss: bool = True,
        reference_id: int = 0,
    ) -> None:
        if penalty_fn is None:
            raise ValueError("Specify penalty function.")
        else:
            if callable(penalty_fn):
                penalty_fn = [penalty_fn]
            self.penalty_fn = penalty_fn

        if prox_penalty is None:
            raise ValueError("Specify proximal operator of penalty function.")
        else:
            if callable(prox_penalty):
                prox_penalty = [prox_penalty]
            self.prox_penalty = prox_penalty

        assert len(self.penalty_fn) == len(
            self.prox_penalty
        ), "Length of penalty_fn and prox_penalty are different."

        if callbacks is not None:
            if callable(callbacks):
                callbacks = [callbacks]
            self.callbacks = callbacks
        else:
            self.callbacks = None

        self.input = None
        self.scale_restoration = scale_restoration

        if reference_id is None and scale_restoration:
            raise ValueError("Specify 'reference_id' if scale_restoration=True.")
        else:
            self.reference_id = reference_id

        self.record_loss = record_loss

        if self.record_loss:
            self.loss = []
        else:
            self.loss = None

    def __call__(self, input: np.ndarray, n_iter: int = 100, **kwargs) -> np.ndarray:
        r"""Separate a frequency-domain multichannel signal.

        Args:
            input (numpy.ndarray):
                Mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            n_iter (int):
                Number of iterations of demixing filter updates.
                Default: ``100``.

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_channels, n_bins, n_frames).
        """
        self.input = input.copy()

        self._reset(**kwargs)

        raise NotImplementedError("Implement '__call__' method.")

    def __repr__(self) -> str:
        s = "PDSBSS("
        s += "n_penalties={n_penalties}".format(n_penalties=self.n_penalties)
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def _reset(self, **kwargs) -> None:
        r"""Reset attributes by given keyword arguments.

        Args:
            kwargs:
                Keyword arguments to set as attributes of PDSBSSBase.
        """
        assert self.input is not None, "Specify data!"

        for key in kwargs.keys():
            setattr(self, key, kwargs[key])

        X = self.input

        n_channels, n_bins, n_frames = X.shape
        n_sources = n_channels  # n_channels == n_sources

        self.n_sources, self.n_channels = n_sources, n_channels
        self.n_bins, self.n_frames = n_bins, n_frames

        if not hasattr(self, "demix_filter"):
            W = np.eye(n_sources, n_channels, dtype=np.complex128)
            W = np.tile(W, reps=(n_bins, 1, 1))
        else:
            if self.demix_filter is None:
                W = None
            else:
                # To avoid overwriting ``demix_filter`` given by keyword arguments.
                W = self.demix_filter.copy()

        self.demix_filter = W
        self.output = self.separate(X, demix_filter=W)

    @property
    def n_penalties(self):
        r"""Return number of penalty terms."""
        return len(self.penalty_fn)

    def separate(self, input: np.ndarray, demix_filter: np.ndarray) -> np.ndarray:
        r"""Separate ``input`` using ``demixing_filter``.

        .. math::
            \boldsymbol{y}_{ij}
            = \boldsymbol{W}_{i}\boldsymbol{x}_{ij}

        Args:
            input (numpy.ndarray):
                The mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            demix_filter (numpy.ndarray):
                The demixing filters to separate ``input``.
                The shape is (n_bins, n_sources, n_channels).

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_sources, n_bins, n_frames).
        """
        X, W = input, demix_filter
        Y = W @ X.transpose(1, 0, 2)
        output = Y.transpose(1, 0, 2)

        return output

    def compute_loss(self) -> float:
        r"""Compute loss :math:`\mathcal{L}`.

        Returns:
            Computed loss.
        """
        X, W = self.input, self.demix_filter
        Y = self.separate(X, demix_filter=W)  # (n_sources, n_bins, n_frames)
        logdet = self.compute_logdet(W)  # (n_bins,)
        penalty = 0

        for penalty_fn in self.penalty_fn:
            penalty = penalty + penalty_fn(Y)

        loss = penalty - np.sum(logdet, axis=0)
        loss = loss.item()

        return loss

    def compute_logdet(self, demix_filter: np.ndarray) -> np.ndarray:
        r"""Compute log-determinant of demixing filter

        Args:
            demix_filter (numpy.ndarray):
                Demixing filters with shape of (n_bins, n_sources, n_channels).

        Returns:
            numpy.ndarray of computed log-determinant values.
        """
        _, logdet = np.linalg.slogdet(demix_filter)  # (n_bins,)

        return logdet

    def normalize_by_spectral_norm(self, input: np.ndarray, n_penalties: int = None) -> np.ndarray:
        r"""Spectral normalization.

        Args:
            input (numpy.ndarray):
                Input spectrogram with shape of (n_channels, n_bins, n_frames).
            n_penalties (int):
                Number of penalty functions, which determines coefficient of normalization.

        Returns:
            numpy.ndarray of normalized spectrogram with shape of (n_channels, n_bins, n_frames).
        """
        if n_penalties is None:
            n_penalties = self.n_penalties

        norm = np.linalg.norm(input.transpose(1, 0, 2), ord=2, axis=(-2, -1))
        norm = np.max(norm)

        return input / (np.sqrt(n_penalties) * norm)

    def restore_scale(self) -> None:
        r"""Restore scale ambiguity.

        If ``self.scale_restoration=projection_back``, we use projection back technique.
        """
        scale_restoration = self.scale_restoration

        assert scale_restoration, "Set self.scale_restoration=True."

        if type(scale_restoration) is bool:
            scale_restoration = "projection_back"

        if scale_restoration in PROJECTION_BACK_KEYWORDS:
            self.apply_projection_back()
        elif scale_restoration in MINIMAL_DISTORTION_PRINCIPLE_KEYWORDS:
            self.apply_minimal_distortion_principle()
        else:
            raise ValueError("{} is not supported for scale restoration.".format(scale_restoration))

    def apply_projection_back(self) -> None:
        r"""Apply projection back technique to estimated spectrograms."""
        assert self.scale_restoration, "Set self.scale_restoration=True."

        X, W = self.input, self.demix_filter
        W_scaled = projection_back(W, reference_id=self.reference_id)
        Y_scaled = self.separate(X, demix_filter=W_scaled)

        self.output, self.demix_filter = Y_scaled, W_scaled

    def apply_minimal_distortion_principle(self) -> None:
        r"""Apply minimal distortion principle to estimated spectrograms."""
        assert self.scale_restoration, "Set self.scale_restoration=True."

        X, W = self.input, self.demix_filter
        Y = self.separate(X, demix_filter=W)
        Y_scaled = minimal_distortion_principle(Y, reference=X, reference_id=self.reference_id)
        X = X.transpose(1, 0, 2)
        Y = Y_scaled.transpose(1, 0, 2)
        X_Hermite = X.transpose(0, 2, 1).conj()
        W_scaled = Y @ X_Hermite @ np.linalg.inv(X @ X_Hermite)

        self.output, self.demix_filter = Y_scaled, W_scaled


class PDSBSS(PDSBSSBase):
    r"""Blind source separation via proximal splitting algorithm [#yatabe2018determined]_.

    Args:
        mu1 (float):
            Step size. Default: ``1``.
        mu2 (float):
            Step size. Default: ``1``.
        alpha (float):
            Step size. Default: ``1``.
        penalty_fn (callable):
            Penalty function that determines source model.
        prox_penalty (callable):
            Proximal operator of penalty function.
            Default: ``None``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back`` explicitly.
            Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back.
            Default: ``0``.
    """

    def __init__(
        self,
        mu1: float = 1,
        mu2: float = 1,
        alpha: float = 1,
        penalty_fn: Callable[[np.ndarray, np.ndarray], float] = None,
        prox_penalty: Callable[[np.ndarray, float], np.ndarray] = None,
        callbacks: Optional[
            Union[Callable[["PDSBSS"], None], List[Callable[["PDSBSS"], None]]]
        ] = None,
        scale_restoration: bool = True,
        record_loss: bool = True,
        reference_id: int = 0,
    ) -> None:
        super().__init__(
            penalty_fn=penalty_fn,
            prox_penalty=prox_penalty,
            callbacks=callbacks,
            scale_restoration=scale_restoration,
            record_loss=record_loss,
            reference_id=reference_id,
        )

        self.mu1, self.mu2 = mu1, mu2
        self.alpha = alpha

    def __call__(self, input, n_iter=100, **kwargs) -> np.ndarray:
        r"""Separate a frequency-domain multichannel signal.

        Args:
            input (numpy.ndarray):
                Mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            n_iter (int):
                Number of iterations of demixing filter updates.
                Default: ``100``.

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_channels, n_bins, n_frames).
        """
        self.input = input.copy()

        self._reset(**kwargs)

        if self.record_loss:
            loss = self.compute_loss()
            self.loss.append(loss)

        if self.callbacks is not None:
            for callback in self.callbacks:
                callback(self)

        for _ in range(n_iter):
            self.update_once()

            if self.record_loss:
                loss = self.compute_loss()
                self.loss.append(loss)

            if self.callbacks is not None:
                for callback in self.callbacks:
                    callback(self)

        if self.scale_restoration:
            self.restore_scale()

        self.output = self.separate(self.input, demix_filter=self.demix_filter)

        return self.output

    def __repr__(self) -> str:
        s = "PDSBSS("
        s += "mu1={mu1}, mu2={mu2}"
        s += ", alpha={alpha}"
        s += ", n_penalties={n_penalties}".format(n_penalties=self.n_penalties)
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def _reset(self, **kwargs) -> None:
        r"""Reset attributes by given keyword arguments.

        Args:
            kwargs:
                Keyword arguments to set as attributes of PDSBSS.
        """
        super()._reset(**kwargs)

        n_penalties = self.n_penalties
        n_sources = self.n_sources
        n_bins, n_frames = self.n_bins, self.n_frames

        if not hasattr(self, "dual"):
            dual = np.zeros((n_penalties, n_sources, n_bins, n_frames), dtype=np.complex128)
        else:
            if self.dual is None:
                dual = None
            else:
                # To avoid overwriting ``dual`` given by keyword arguments.
                dual = self.dual.copy()

        self.dual = dual

    def update_once(self) -> None:
        r"""Update demixing filters and dual parameters once."""
        mu1, mu2 = self.mu1, self.mu2
        alpha = self.alpha

        Y = self.dual
        X, W = self.input, self.demix_filter

        Y_sum = Y.sum(axis=0)
        XY = Y_sum.transpose(1, 0, 2) @ X.transpose(1, 2, 0).conj()
        W_tilde = prox.neg_logdet(W - mu1 * mu2 * XY, step_size=mu1)
        XW = self.separate(X, demix_filter=2 * W_tilde - W)
        Y_tilde = []

        for Y_q, prox_penalty in zip(Y, self.prox_penalty):
            Z_q = Y_q + XW
            Y_tilde_q = Z_q - prox_penalty(Z_q, step_size=1 / mu2)
            Y_tilde.append(Y_tilde_q)

        Y_tilde = np.stack(Y_tilde, axis=0)

        self.demix_filter = alpha * W_tilde + (1 - alpha) * W
        self.dual = alpha * Y_tilde + (1 - alpha) * Y



from typing import Callable, Iterable, Optional, Tuple

import numpy as np
import pytest

from ssspy.bss._update_spatial_model import (
    _psd_inv,
    update_by_block_decomposition_vcd,
    update_by_ip1,
    update_by_ip2,
    update_by_ip2_one_pair,
    update_by_iss1,
    update_by_iss2,
)
from ssspy.special import add_flooring, max_flooring
from ssspy.utils.select_pair import combination_pair_selector, sequential_pair_selector


def negative_pair_selector(n_sources):
    for m in range(n_sources):
        m, n = m % n_sources, (m + 1) % n_sources
        m, n = m - n_sources, n - n_sources

        yield m, n


parameters = [(31, 20)]
parameters_block_decomposition_vcd = [(15, 2, 20)]
parameters_n_sources = [2, 3]
parameters_flooring_fn = [max_flooring, add_flooring, None]
parameters_overwrite = [True, False]
parameters_singular_fn = [
    lambda x: np.abs(x) < max_flooring(x),
    lambda x: np.abs(x) < add_flooring(x),
    None,
]
parameters_pair_selector = [
    sequential_pair_selector,
    combination_pair_selector,
    negative_pair_selector,
    None,
]


@pytest.mark.parametrize("n_bins, n_frames", parameters)
@pytest.mark.parametrize("n_sources", parameters_n_sources)
@pytest.mark.parametrize("flooring_fn", parameters_flooring_fn)
def test_update_by_ip1(
    n_bins: int,
    n_frames: int,
    n_sources: int,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]],
):
    n_channels = n_sources

    rng = np.random.default_rng(42)

    varphi = 1 / rng.random((n_sources, n_frames))
    X = rng.standard_normal((n_channels, n_bins, n_frames))
    real = rng.standard_normal((n_bins, n_sources, n_sources))
    imag = rng.standard_normal((n_bins, n_sources, n_sources))
    W = real + 1j * imag

    XX_Hermite = X[:, np.newaxis, :, :] * X[np.newaxis, :, :, :].conj()
    XX_Hermite = XX_Hermite.transpose(2, 0, 1, 3)
    GXX = varphi[:, np.newaxis, np.newaxis, :] * XX_Hermite[:, np.newaxis, :, :, :]
    U = np.mean(GXX, axis=-1)
    W_updated = update_by_ip1(W, U, flooring_fn=flooring_fn)

    assert W_updated.shape == W.shape


@pytest.mark.parametrize("n_bins, n_frames", parameters)
@pytest.mark.parametrize("n_sources", parameters_n_sources)
@pytest.mark.parametrize("flooring_fn", parameters_flooring_fn)
@pytest.mark.parametrize("pair_selector", parameters_pair_selector)
def test_update_by_ip2(
    n_bins: int,
    n_frames: int,
    n_sources: int,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]],
    pair_selector: Optional[Callable[[int], Iterable[Tuple[int, int]]]],
):
    n_channels = n_sources

    rng = np.random.default_rng(42)

    varphi = 1 / rng.random((n_sources, n_frames))
    X = rng.standard_normal((n_channels, n_bins, n_frames))
    real = rng.standard_normal((n_bins, n_sources, n_sources))
    imag = rng.standard_normal((n_bins, n_sources, n_sources))
    W = real + 1j * imag

    XX_Hermite = X[:, np.newaxis, :, :] * X[np.newaxis, :, :, :].conj()
    XX_Hermite = XX_Hermite.transpose(2, 0, 1, 3)
    GXX = varphi[:, np.newaxis, np.newaxis, :] * XX_Hermite[:, np.newaxis, :, :, :]
    U = np.mean(GXX, axis=-1)
    W_updated = update_by_ip2(W, U, flooring_fn=flooring_fn, pair_selector=pair_selector)

    assert W_updated.shape == W.shape


@pytest.mark.parametrize("n_bins, n_frames", parameters)
@pytest.mark.parametrize("n_sources", parameters_n_sources)
@pytest.mark.parametrize("flooring_fn", parameters_flooring_fn)
def test_update_by_ip2_one_pair(
    n_bins: int,
    n_frames: int,
    n_sources: int,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]],
):
    n_channels = n_sources

    rng = np.random.default_rng(42)

    varphi = 1 / rng.random((2, n_bins, n_frames))
    X = rng.standard_normal((n_channels, n_bins, n_frames))
    real = rng.standard_normal((n_bins, n_sources, n_channels))
    imag = rng.standard_normal((n_bins, n_sources, n_channels))
    W = real + 1j * imag
    XX = X[:, np.newaxis] * X[np.newaxis, :].conj()
    GXX = np.mean(varphi[:, np.newaxis, np.newaxis, :, :] * XX[np.newaxis, :, :, :, :], axis=-1)
    GXX = GXX.transpose(3, 0, 1, 2)

    W_updated = update_by_ip2_one_pair(W, GXX, pair=(1, 0), flooring_fn=flooring_fn)

    assert W_updated.shape == (n_bins, 2, n_channels)


@pytest.mark.parametrize("n_bins, n_frames", parameters)
@pytest.mark.parametrize("n_sources", parameters_n_sources)
@pytest.mark.parametrize("flooring_fn", parameters_flooring_fn)
def test_update_by_iss1(
    n_bins: int,
    n_frames: int,
    n_sources: int,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]],
):
    rng = np.random.default_rng(42)

    varphi = 1 / rng.random((n_sources, n_bins, n_frames))
    real = rng.standard_normal((n_sources, n_bins, n_frames))
    imag = rng.standard_normal((n_sources, n_bins, n_frames))
    Y = real + 1j * imag

    Y_updated = update_by_iss1(Y, varphi, flooring_fn=flooring_fn)

    assert Y_updated.shape == Y.shape


@pytest.mark.parametrize("n_bins, n_frames", parameters)
@pytest.mark.parametrize("n_sources", parameters_n_sources)
@pytest.mark.parametrize("flooring_fn", parameters_flooring_fn)
@pytest.mark.parametrize("pair_selector", parameters_pair_selector)
def test_update_by_iss2(
    n_bins: int,
    n_frames: int,
    n_sources: int,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]],
    pair_selector: Optional[Callable[[int], Iterable[Tuple[int, int]]]],
):
    rng = np.random.default_rng(42)

    varphi = 1 / rng.random((n_sources, n_bins, n_frames))
    real = rng.standard_normal((n_sources, n_bins, n_frames))
    imag = rng.standard_normal((n_sources, n_bins, n_frames))
    Y = real + 1j * imag

    Y_updated = update_by_iss2(Y, varphi, flooring_fn=flooring_fn, pair_selector=pair_selector)

    assert Y_updated.shape == Y.shape


@pytest.mark.parametrize("n_blocks, n_neighbors, n_frames", parameters_block_decomposition_vcd)
@pytest.mark.parametrize("n_sources", parameters_n_sources)
@pytest.mark.parametrize("singular_fn", parameters_singular_fn)
@pytest.mark.parametrize("overwrite", parameters_overwrite)
def test_update_by_block_decomposition_vcd(
    n_blocks: int,
    n_neighbors: int,
    n_frames: int,
    n_sources: int,
    singular_fn: Optional[Callable[[np.ndarray], np.ndarray]],
    overwrite: bool,
):
    na = np.newaxis
    n_channels = n_sources

    rng = np.random.default_rng(42)

    R = rng.random((n_blocks, n_neighbors, n_sources, n_channels, n_frames))
    X = rng.standard_normal((n_channels, n_blocks, n_neighbors, n_frames))
    real = rng.standard_normal((n_blocks, n_neighbors, n_sources, n_channels))
    imag = rng.standard_normal((n_blocks, n_neighbors, n_sources, n_channels))
    W = real + 1j * imag

    R = R[:, :, na, :, :, :] * np.eye(n_neighbors)[:, :, na, na, na]
    R = R[:, :, :, :, :, na, :] * np.eye(n_channels)[:, :, na]

    XX_Hermite = X[:, na, :, :, na] * X[na, :, :, na, :].conj()
    XX_Hermite = XX_Hermite.transpose(2, 3, 4, 0, 1, 5)

    RXX = np.mean(R * XX_Hermite[:, :, :, na], axis=-1)

    W_updated = update_by_block_decomposition_vcd(
        W, weighted_covariance=RXX, singular_fn=singular_fn, overwrite=overwrite
    )

    assert W_updated.shape == W.shape


def test_psd_inv() -> None:
    rng = np.random.default_rng(42)

    n_bins, n_frames = 129, 100
    n_sources = n_channels = 4

    varphi = 1 / rng.random((n_sources, n_frames))
    X = rng.standard_normal((n_channels, n_bins, n_frames))

    XX_Hermite = X[:, np.newaxis, :, :] * X[np.newaxis, :, :, :].conj()
    XX_Hermite = XX_Hermite.transpose(2, 0, 1, 3)
    GXX = varphi[:, np.newaxis, np.newaxis, :] * XX_Hermite[:, np.newaxis, :, :, :]
    U = np.mean(GXX, axis=-1)

    U_inv = _psd_inv(U)
    eye = np.eye(n_sources)

    assert np.allclose(U @ U_inv, eye)



# It is expected to run from root ssspy directory
from tests.dummy.utils.dataset import download_sample_speech_data


def download_all() -> None:
    conditions = [
        {"n_sources": 2, "sisec2010_tag": "dev1_female3"},
        {"n_sources": 3, "sisec2010_tag": "dev1_female3"},
        {"n_sources": 4, "sisec2010_tag": "dev1_female4"},
    ]
    max_durations = [0.1, 0.5]

    for kwargs in conditions:
        for max_duration in max_durations:
            download_sample_speech_data(max_duration=max_duration, **kwargs)


if __name__ == "__main__":
    download_all()



from typing import Optional

import numpy as np


def projection_back(
    data_or_filter: np.ndarray,
    reference: Optional[np.ndarray] = None,
    reference_id: Optional[int] = 0,
) -> np.ndarray:
    r"""Projection back technique to restore scale ambiguity.

    The implementation is based on [#murata2001approach]_.

    Args:
        data_or_filter (numpy.ndarray):
            Estimated spectrograms or demixing filters.
        reference (numpy.ndarray, optional):
            Reference spectrogram.
        reference_id (int, optional):
            Reference microphone index. Default: ``0``.

    Returns:
        numpy.ndarray of rescaled estimated spectrograms or demixing filters.

    Examples:
        When you give estimated spectrograms,

        .. code-block:: python

            >>> import numpy as np
            >>> from ssspy.algorithm import projection_back

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> n_sources = n_channels
            >>> rng = np.random.default_rng(42)

            >>> spectrogram_mix = \
            ...     rng.standard_normal((n_channels, n_bins, n_frames)) \
            ...     + 1j * rng.standard_normal((n_channels, n_bins, n_frames))
            >>> demix_filter = \
            ...     rng.standard_normal((n_sources, n_channels)) \
            ...     + 1j * rng.standard_normal((n_sources, n_channels))

            >>> spectrogram_est = demix_filter @ spectrogram_mix.transpose(1, 0, 2)

            >>> # (n_bins, n_sources, n_frames) -> (n_sources, n_bins, n_frames)
            >>> spectrogram_est = spectrogram_est.transpose(1, 0, 2)

            >>> spectrogram_est_scaled = \
            ...     projection_back(spectrogram_est, reference=spectrogram_mix, reference_id=0)
            >>> spectrogram_est_scaled.shape
            (2, 2049, 128)

        When you give demixing filters,

        .. code-block:: python

            >>> import numpy as np
            >>> from ssspy.algorithm import projection_back

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> n_sources = n_channels
            >>> rng = np.random.default_rng(42)

            >>> spectrogram_mix = \
            ...     rng.standard_normal((n_channels, n_bins, n_frames)) \
            ...     + 1j * rng.standard_normal((n_channels, n_bins, n_frames))
            >>> demix_filter = \
            ...     rng.standard_normal((n_sources, n_channels)) \
            ...     + 1j * rng.standard_normal((n_sources, n_channels))

            >>> demix_filter_scaled = projection_back(demix_filter, reference_id=0)

            >>> spectrogram_est_scaled = demix_filter_scaled @ spectrogram_mix.transpose(1, 0, 2)

            >>> # (n_bins, n_sources, n_frames) -> (n_sources, n_bins, n_frames)
            >>> spectrogram_est_scaled = spectrogram_est_scaled.transpose(1, 0, 2)
            >>> spectrogram_est_scaled.shape
            (2, 2049, 128)

    .. [#murata2001approach]
        N. Murata, S. Ikeda, and A. Ziehe,
        "An approach to blind source separation based on temporal structure of speech signals,"
        *Neurocomputing*, vol. 41, no. 1-4, pp. 1-24, 2001.
    """
    if reference is None:
        W = data_or_filter  # (*, n_sources, n_channels)
        scale = np.linalg.inv(W)  # (*, n_channels, n_sources)

        if reference_id is None:
            scale = scale[..., np.newaxis]  # (*, n_channels, n_sources, 1)
            scale = np.rollaxis(scale, -3, 0)  # (n_channels, *, n_sources, 1)
            demix_filter_scaled = W * scale  # (n_channels, *, n_sources, n_channels)
        else:
            scale = scale[..., reference_id, :]  # (*, n_sources)
            demix_filter_scaled = W * scale[..., np.newaxis]  # (*, n_sources, n_channels)

        return demix_filter_scaled
    else:
        Y = data_or_filter  # (n_sources, n_bins, n_frames)
        X = reference  # (n_channels, n_bins, n_frames)

        Y = Y.transpose(1, 0, 2)  # (n_bins, n_sources, n_frames)
        X = X.transpose(1, 0, 2)  # (n_bins, n_channels, n_frames)
        Y_Hermite = Y.transpose(0, 2, 1).conj()  # (n_bins, n_frames, n_sources)
        XY_Hermite = X @ Y_Hermite  # (n_bins, n_channels, n_sources)
        YY_Hermite = Y @ Y_Hermite  # (n_bins, n_sources, n_sources)

        scale = XY_Hermite @ np.linalg.inv(YY_Hermite)  # (n_bins, n_channels, n_sources)

        if reference_id is None:
            scale = scale.transpose(1, 0, 2)  # (n_channels, n_bins, n_sources)
            Y_scaled = Y * scale[..., np.newaxis]  # (n_channels, n_bins, n_sources, n_frames)
            output_scaled = Y_scaled.swapaxes(-3, -2)  # (n_channels, n_sources, n_bins, n_frames)
        else:
            scale = scale[..., reference_id, :]  # (n_bins, n_sources)
            Y_scaled = Y * scale[..., np.newaxis]  # (n_bins, n_sources, n_frames)
            output_scaled = Y_scaled.swapaxes(-3, -2)  # (n_sources, n_bins, n_frames)

        return output_scaled



import os
import sys
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pytest
import scipy.signal as ss

from ssspy.bss.pdsbss import PDSBSS, PDSBSSBase

ssspy_tests_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
sys.path.append(ssspy_tests_dir)

from dummy.callback import DummyCallback, dummy_function  # noqa: E402
from dummy.utils.dataset import download_sample_speech_data  # noqa: E402

max_duration = 0.5
n_fft = 2048
hop_length = 1024
n_bins = n_fft // 2 + 1
n_iter = 5

parameters_pdsbss = [
    (2, None, {}),
    (
        3,
        dummy_function,
        {"demix_filter": np.tile(-np.eye(3, dtype=np.complex128), reps=(n_bins, 1, 1))},
    ),
    (2, [DummyCallback(), dummy_function], {}),
]


def contrast_fn(y: np.ndarray) -> np.ndarray:
    r"""Contrast function.

    Args:
        y (np.ndarray):
            The shape is (n_sources, n_bins, n_frames).

    Returns:
        np.ndarray of the shape is (n_sources, n_frames).
    """
    return 2 * np.linalg.norm(y, axis=1)


def penalty_fn(y: np.ndarray) -> float:
    loss = contrast_fn(y)
    loss = np.sum(loss.mean(axis=-1))
    return loss


def prox_penalty(y: np.ndarray, step_size: float = 1) -> np.ndarray:
    r"""Proximal operator of penalty function.

    Args:
        y (np.ndarray):
            The shape is (n_sources, n_bins, n_frames).
        step_size (float):
            Step size. Default: 1.

    Returns:
        np.ndarray of the shape is (n_sources, n_bins, n_frames).
    """
    norm = np.linalg.norm(y, axis=1, keepdims=True)
    return y * np.maximum(1 - step_size / norm, 0)


def test_pds_base():
    class DummyPDSBSS(PDSBSSBase):
        pass

    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        sisec2010_root="./tests/.data/SiSEC2010",
        mird_root="./tests/.data/MIRD",
        n_sources=2,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    pdsbss = DummyPDSBSS(penalty_fn=penalty_fn, prox_penalty=prox_penalty)

    with pytest.raises(NotImplementedError) as e:
        _ = pdsbss(spectrogram_mix, n_iter=n_iter)

    assert str(e.value) == "Implement '__call__' method."

    print(pdsbss)


@pytest.mark.parametrize("n_sources, callbacks, reset_kwargs", parameters_pdsbss)
def test_pdsbss(
    n_sources: int,
    callbacks: Optional[Union[Callable[[PDSBSS], None], List[Callable[[PDSBSS], None]]]],
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        sisec2010_root="./tests/.data/SiSEC2010",
        mird_root="./tests/.data/MIRD",
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    pdsbss = PDSBSS(penalty_fn=penalty_fn, prox_penalty=prox_penalty, callbacks=callbacks)
    spectrogram_mix_normalized = pdsbss.normalize_by_spectral_norm(spectrogram_mix)
    spectrogram_est = pdsbss(spectrogram_mix_normalized, n_iter=n_iter, **reset_kwargs)

    assert spectrogram_mix.shape == spectrogram_est.shape

    print(pdsbss)
